/****************************************************************************
  kevobot -- a Khepera simulator for Webots with kevopic functionalities
  Copyright (C) 2003 Autonomous Systems Lab 2, EPFL, Lausanne

  This program is free software; the authors of any publication 
  arising from research using this software are asked to add the 
  following reference:
    kevobot; http://asl.epfl.ch/resources/evo/kevobot.php; 
    Autonomous Systems Lab 2, EPFL, Lausanne.

  You can redistribute this program and/or modify it under the 
  terms of the GNU General Public License as published by the 
  Free Software Foundation; either version 2 of the License, 
  or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
******************************************************************************/

/**********************************************************************************************************************

  Authors:      Jean-Christophe Zufferey (jcz)                                                                                                              
 
  History: 
		01.01.03 (jcz)		Removed the option for more than one camera
					 						Corrected FOV for K213
					 						Dynamic pixel number: 16 or 64
		23.11.02 (jcz)		Adapted from Blimp, Khepera_K213, etc.
		25.11.02 (jcz)		Divers bug fixes. Added encoders.
		27.11.02 (jcz)		Some corrections on the encoders & noise + scaling added to vision.
		24.03.03 (jcz)		Addapted to KEvoPic code. Floor sensor added.
											Partially removed glib dependencies in anticipation of Webots 4
	  26.03.03 (jcz)		Added support for both K213 (virtual) and TSL3301 :
											K213 and TSL3301 are now two distinct camera on the simulated robot
											with a more accurate position and FOV.
    26.03.03 (jcz)    Last modifications for Webots 4 compatibility.
    10.05.03 (jcz)		Added 0x3B command (khep_read_delta_position) and VIRTUAL_WALLS
    									Encoder values have been adjusted to be a more realistic.
		03.07.03 (jcz)		Handling of disconnection from i.
		21.10.03 (jcz)		VIRTUAL WALL is always enabled but not active by default; command 0x20 must be used
		
	Todo:
		- Double-check TSL3301 FOV precision
		- Implement commands 0x2A & 0x2B
		- Implement command 0x11 (light integration time)?

**********************************************************************************************************************/

//---------------------------------------------------------------------------------------------------------------
// Important settings. Some of them must match parameters in goevo at compile time.
//---------------------------------------------------------------------------------------------------------------

// Cycle time
#define Dt  0.100       		// time step [s]	(MUST BE IDENTICAL TO SETTINGS IN i)

// Vision
#define CAMERA_TYPE	1				// 0 is K213 with 16px (36 FOV), 1 is TSL3301 with 25px (70 FOV)	(MUST BE IDENTICAL TO SETTINGS IN i)
#define VISION_NOISE 15			// ~ +- 10% on 256
#define VISION_SCALING 0.5	// toward white (sensor saturation is more probable than absolutely no light)

// Virtual walls
#define VW_THRESHOLD	250		// on 8-bit (not 10-bit)

//---------------------------------------------------------------------------------------------------------------

#include <stdio.h>
#include <string.h>

// Utils
#include <stdlib.h>
float rand_float(float a);

// TCPIP
#ifdef WIN32
#include <winsock.h>
#else
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h> /* definition of struct sockaddr_in */
#include <netdb.h>      /* definition of gethostbyname */
#include <arpa/inet.h>  /* definition of inet_ntoa */
#include <unistd.h>     /* definition of close */
#endif

#define SOCKET_PORT 10020

int accept_client(int server_fd);
int create_socket_server(int port);
int close_connection(int server_fd);

// Webots stuffs
#include <device/robot.h>
#include <device/camera.h>
#include <device/differential_wheels.h>
#include <device/distance_sensor.h>

#include <math.h>

// Distance sensors
#define NB_DIST_SENS 8
static DeviceTag ds[NB_DIST_SENS];
unsigned char ds_value[NB_DIST_SENS];

// Floor sensor
static DeviceTag fs;

// Vision
DeviceTag camera;

	// K213	
#if CAMERA_TYPE == 0	
#define NUM_PIXELS 16
#define CAMERA_FOV 0.63 		// K213 has 36 FOV => 0.63 rad
#define IMAGE_WIDTH 64			// Defined in the Webots World
#define IMAGE_HEIGHT IMAGE_WIDTH
	// TSL3301
#elif CAMERA_TYPE == 1  		
#define NUM_PIXELS 25
#define CAMERA_FOV 1.2 			// 70 FOV => 1.2 rad
#define IMAGE_WIDTH 50			// Defined in the Webots World
#define IMAGE_HEIGHT IMAGE_WIDTH
#endif

#define VISUALSCALE 255.0

// Encoders
int left_encoder, right_encoder;
int left_speed, right_speed;


void reset(void) 
{
	char text[4];
  int i;
  
	printf("\n--- RESET ---\n"); 
	printf("Robot name is %s\n", robot_get_name());
  
  text[1]='s';
  text[3]='\0';
  for (i=0;i<NB_DIST_SENS;i++) {
    text[0]='d';
    text[2]='0'+i;
    ds[i] = robot_get_device(text); /* distance sensors */
    if (ds[i]==0) printf("Distance sensor %d is missing!\n", i);
  } 
  
  fs = robot_get_device("fs0"); /* floor sensor */
  if (fs==0) printf("Floor sensor is missing!\n");

#if CAMERA_TYPE == 0	
  camera = robot_get_device("K213");
  if (camera==0) printf("K213 is missing!\n");
#elif CAMERA_TYPE == 1 
  camera = robot_get_device("TSL3301");
  if (camera==0) printf("TSL3301 is missing!\n");
#endif

}

int main()
{
	int i;
	int temp;
	float val;
	short unsigned int virtual_walls_enabled=0;

	// TCPIP
	int fd;
	unsigned char RxBuffer[256], TxBuffer[256];

	// Camera
	unsigned char *h_image;
	float mean, greyvalue;
	
	printf("\n********************************************************************\n\n");
  printf("  Set the Basic Simulation Step to 50ms in Webots preferences.\n\n");
  printf("  Make sure the following parameters are set correctly in goevo:\n");
  printf("    - simulated cycle time: %d ms\n", (int)(Dt*1000));
#if CAMERA_TYPE == 0	
  printf("    - camera type: K213 with 16 pixels and 36 deg FOV\n");
#elif CAMERA_TYPE == 1 
  printf("    - camera type: TSL3301 with 25 pixels and 70 deg FOV\n");
#endif
	printf("\n********************************************************************\n\n");

	robot_live(reset);

	camera_enable(camera,(short unsigned int)(Dt*1000));
	camera_set_fov(camera, CAMERA_FOV);

	for(i=0;i<NB_DIST_SENS;i++) distance_sensor_enable(ds[i],(short unsigned int)(Dt*1000));
	distance_sensor_enable(fs,(short unsigned int)(Dt*1000));
	
	differential_wheels_enable_encoders((short unsigned int)(Dt*1000));

	fd = create_socket_server(SOCKET_PORT);	// waiting for connection from i

	for(;;) { /* The robot never dies! */
	
		// TCPIP communication
receive_packet:
  	while (recv(fd,RxBuffer,2,0)<0);    // waiting for the first byte of the packet

		// Handle disconnection packet
		if (RxBuffer[0]==0xFF && RxBuffer[1]==0xFF) {
			close_connection(fd);
   		printf("Disconnected.\n\n");
   		fd = create_socket_server(SOCKET_PORT);	// waiting for connection from i
   		goto receive_packet;
		}
		
		switch(RxBuffer[0]) {	
	
			case 0x01:
				TxBuffer[0]=RxBuffer[0];
				TxBuffer[1]=0;
				while (recv(fd,RxBuffer,1,0)<0);
				if (RxBuffer[0] & 0x01) { 
    			virtual_walls_enabled = 1;
       		printf("VIRTUAL WALL activated.\n");
     		} else {
     			virtual_walls_enabled = 0; 
        	printf("VIRTUAL WALL deactivated.\n");
       	}
				send(fd,TxBuffer,2,0);
				break;			

#if CAMERA_TYPE == 1
			// TSL3301
			case 0x10:  // grab and send image	
				TxBuffer[0]=RxBuffer[0];
				TxBuffer[1]=NUM_PIXELS;
				send(fd,TxBuffer,2,0);
		  	h_image = camera_get_image(camera);
		  	// The image from webots is 64x64px. We start at pixel (64-50)/2 = 7 
     		// and then read 1 pixel every 2 on the line at the middle of the height.
   			for (i=0; i<NUM_PIXELS; i++) {
 					mean = 0;
        	greyvalue  = camera_image_get_red(h_image, IMAGE_WIDTH, i*(IMAGE_WIDTH/NUM_PIXELS), IMAGE_HEIGHT/2);
        	greyvalue += camera_image_get_green(h_image, IMAGE_WIDTH, i*(IMAGE_WIDTH/NUM_PIXELS), IMAGE_HEIGHT/2); 
        	greyvalue += camera_image_get_blue(h_image, IMAGE_WIDTH, i*(IMAGE_WIDTH/NUM_PIXELS), IMAGE_HEIGHT/2); 
					mean = (greyvalue/3)+rand_float(VISION_NOISE);				// add noise
					mean = mean*VISION_SCALING + (1-VISION_SCALING)*255;	// scale image
					if (mean<0) mean=0;
					else if (mean>255) mean=255;			
        	TxBuffer[i] = (char)(mean);
   			}	  
		  	send(fd,TxBuffer,NUM_PIXELS,0);
				break;
			case 0x11:	// set the light integration time (not implemented)
				TxBuffer[0]=RxBuffer[0];
				TxBuffer[1]=0;
				while (recv(fd,RxBuffer,1,0)<0);	// 1 parameter: exposition time in 100um
				send(fd,TxBuffer,2,0);
				break;
#endif
				
#if CAMERA_TYPE == 0				
			// Virtual K213
			case 0x1A:  // grab and send image	
				TxBuffer[0]=RxBuffer[0];
				TxBuffer[1]=NUM_PIXELS;
				send(fd,TxBuffer,2,0);
		  	h_image = camera_get_image(camera);
   			for (i=0; i<NUM_PIXELS; i++) {
 					mean = 0;
        	greyvalue  = camera_image_get_red(h_image, IMAGE_WIDTH, i*(IMAGE_WIDTH/NUM_PIXELS), IMAGE_HEIGHT/2);
        	greyvalue += camera_image_get_green(h_image, IMAGE_WIDTH, i*(IMAGE_WIDTH/NUM_PIXELS), IMAGE_HEIGHT/2); 
        	greyvalue += camera_image_get_blue(h_image, IMAGE_WIDTH, i*(IMAGE_WIDTH/NUM_PIXELS), IMAGE_HEIGHT/2); 
					mean = (greyvalue/3)+rand_float(VISION_NOISE);
					mean = mean*VISION_SCALING + (1-VISION_SCALING)*255;
					if (mean<0) mean=0;
					else if (mean>255) mean=255;			
        	TxBuffer[i] = (char)(mean);
   			}	  
		  	send(fd,TxBuffer,NUM_PIXELS,0);
				break;
#endif
		
			// Basic Khepera
			case 0x20:
				TxBuffer[0]=0x20;
				TxBuffer[1]=NB_DIST_SENS;
				send(fd,TxBuffer,2,0);
				for (i=0; i<NB_DIST_SENS; i++) {
					ds_value[i] = (unsigned char)(distance_sensor_get_value(ds[i])>>2); // 10-bit -> 8-bit
					TxBuffer[i] = ds_value[i]; 
				}
				send(fd,TxBuffer,NB_DIST_SENS,0);
				break;

			// Floor sensor (not yet implemented)
			case 0x25:
				TxBuffer[0]=RxBuffer[0];
				TxBuffer[1]=1;
				temp = distance_sensor_get_value(fs);
				// shift and invert the value in order to be compatible with real robot's floor sensor
				temp = (temp - 124);
				temp /= 4;    // 10-bit -> 8-bit
				TxBuffer[2]=(char)(temp); 
				send(fd,TxBuffer,3,0);
				break;
				
			// Read all sensors (not yet implemented)
			case 0x2A:

				break;

			// Read all sensors but IR (not yet implemented)
			case 0x2B:

				break;
		
			// Wheel speeds (!! should always be called after 0x3A !!)
 			case 0x30:	 // set the speed of the motors 
				while (recv(fd,RxBuffer,2,0)<0);    // waiting for the motor commands
				val = (float)differential_wheels_get_left_encoder();
				left_encoder = (int)(val/2);
				left_speed = (int)(val/Dt/100);	// Dt is [s], speed (val) is in [pulse/10ms], Dt/100 is [10ms]
				val = (float)differential_wheels_get_right_encoder();
				right_encoder = (int)(val/2);
				right_speed = (int)(val/Dt/100);
				// VIRTUAL_WALLS
				if (virtual_walls_enabled==1) {		
					int speedSum = (signed char)RxBuffer[0]+(signed char)RxBuffer[1];
					// Check speeds...
					if (ds_value[1]>VW_THRESHOLD || ds_value[2]>VW_THRESHOLD || ds_value[3]>VW_THRESHOLD || ds_value[4]>VW_THRESHOLD) {
						if (speedSum>0) {
							RxBuffer[0]=0;
							RxBuffer[1]=0;
//							printf("VIRTUAL WALL active (forward).\n");
						} 
					} else if (ds_value[6]>VW_THRESHOLD || ds_value[7]>VW_THRESHOLD) {
						if (speedSum<0) {
							RxBuffer[0]=0;
							RxBuffer[1]=0;
//							printf("VIRTUAL WALLS active (backward).\n");
						}
					}
				}	
				differential_wheels_set_speed((signed char)RxBuffer[0],(signed char)RxBuffer[1]);
				differential_wheels_set_encoders(0,0);  // Reset encoders for next step...
				TxBuffer[0]=0x30;
				TxBuffer[1]=0;
 				send(fd,TxBuffer,2,0);
 				robot_step((int)(Dt*1000)); /* run one simulated step */
				break;
				
			case 0x3A:		/* read wheel encoders */
				TxBuffer[0]=0x3A;
				TxBuffer[1]=2;
				send(fd,TxBuffer,2,0);
				TxBuffer[0]=(short int)left_speed;
				TxBuffer[1]=(short int)right_speed;
 				send(fd,TxBuffer,2,0);
				break;
				
			case 0x3B:		/* read delta positions */
				TxBuffer[0]=0x3B;
				TxBuffer[1]=2;
				send(fd,TxBuffer,2,0);
				TxBuffer[0]=(short int)left_encoder;
				TxBuffer[1]=(short int)right_encoder;
 				send(fd,TxBuffer,2,0);
				break;
         	
     	// NOP
     	default:
				TxBuffer[0]=RxBuffer[0];
				TxBuffer[1]=0;
				send(fd,TxBuffer,2,0);
       	break;
		
		}		// switch
  }
	
	close_connection(fd);

	return 0;
}

//----------------------------------------------------------------------------
// Utilities

// Returns a random float between -a and +a
float rand_float(float a) {
	return 2.0f * a * (rand() / (float)RAND_MAX) - a;
}


//----------------------------------------------------------------------------
// TCP-IP server

int accept_client(int server_fd) {
  int cfd;
  struct sockaddr_in client;
  int asize = sizeof(struct sockaddr_in);
  struct hostent *client_info;

  printf("Waiting for connection...\n");
  cfd = accept(server_fd,(struct sockaddr *)&client,&asize);
  if (cfd==-1) {
    printf("Cannot accept client!\n");
    return -1;
  }
  client_info = gethostbyname((char *)inet_ntoa(client.sin_addr));
  printf("Accepted connection from: %s\n",client_info->h_name);
  return cfd;
}

int create_socket_server(int port) {
  int sfd,rc;
  struct sockaddr_in address;

#ifdef WIN32
  /* initialize the socket api */
  WSADATA info;
  rc = WSAStartup(MAKEWORD(1,1),&info); /* Winsock 1.1 */
  if (rc!=0) {
    printf("Cannot initialize Winsock\n");
    return -1;
  }
#endif
  /* create the socket */
  sfd = socket(AF_INET,SOCK_STREAM,0);
  if (sfd==-1) {
    printf("Cannot create socket\n");
    return -1;
  }
  /* fill in socket address */
  memset(&address,0,sizeof(struct sockaddr_in));
  address.sin_family = AF_INET;
  address.sin_port = htons(port);
  address.sin_addr.s_addr = INADDR_ANY;
  /* bind to port */
  rc = bind(sfd,(struct sockaddr *)&address,sizeof(struct sockaddr));
  if (rc==-1) {
    printf("Cannot bind port %d\n",port);
    return -1;
  }
  /* listen for connections */
  if (listen(sfd,1)==-1) {
    printf("Cannot listen for connections\n");
    return -1;
  }
  printf("Waiting for a connection on port %d...\n",port);
  return accept_client(sfd);
}

int close_connection(int server_fd) {
	int ret;
#ifdef WIN32
	ret = closesocket(server_fd);
#else
	ret = close(server_fd);
#endif
	if (ret!=0) printf("Cannot close socket.\n");
#ifdef WIN32
	WSACleanup();	// Terminate use of the Windows Sockets DLL. 
#endif
	return ret;
}
